"""Factiva Core User Stream Class."""

import json
import pandas as pd

from google.cloud.pubsub_v1 import SubscriberClient
from google.oauth2 import service_account

from factiva.core import const, APIKeyUser
from factiva.helper import api_send_request
from .stream_response import StreamResponse


class StreamUser(APIKeyUser):
    """
    Implements the class that represents an Stream user.

    This entity is an extension from the APIKeyUser class.
    This class is capable of consulting the list streams from a given user
    It can authenticate for obtaining the required credentials for Pubsub
    The authentication can be with OAuth or User Key

    Parameters
    ----------
    api_key : str
        String containing the 32-character long APi Key. If not provided, the
        constructor will try to obtain its value from the FACTIVA_APIKEY
        environment variable.
    request_info : boolean, optional (Default: False)
        Indicates if user data has to be pulled from the server.
        This operation fills account detail properties along with maximum,
        used and remaining values. It may take several seconds to complete.

    See Also
    --------
    UserOAuth : API user that follows the OAuth guidelines.

    Examples
    --------
    Creates a new stream user.
        >>> stream_user = StreamUser(
                api_key='abcd1234abcd1234abcd1234abcd1234',
                request_info=False
            )
            print(stream_user.get_streams())

    """

    __API_ACCOUNT_STREAM_CREDENTIALS_BASEPATH = const.API_ACCOUNT_STREAM_CREDENTIALS_BASEPATH
    __DEFAULT_HOST_DNA = f'{const.API_HOST}{const.DNA_BASEPATH}'
    __DEFAULT_HOST_ALPHA = f'{const.API_HOST}{const.ALPHA_BASEPATH}'
    __API_ENDPOINT_STREAM_URL = f'{const.API_HOST}{const.API_STREAMS_BASEPATH}/'

    def __init__(
        self,
        api_key=None,
        request_info=False,
    ):
        """Construct the object instance."""
        super().__init__(api_key, request_info)

    # TODO: Please remove as this is a functionality that belongs to the News package
    def get_streams(self) -> pd.DataFrame:
        """Obtain streams from a given user.

        Function which returns the streams a given user with
        its respective key using the default stream url

        Returns
        -------
        Json object -> list of objects containing
        information about every stream (id, link, state, etc)

        Raises
        ------
        AttributeError:
            When is not possible to parse the data as json or dataframe
        ValueError:
            When API key is not valid
        RuntimeError:
            When API request returns unexpected error

        """
        request_headers = {'user-key': self.api_key}
        response = api_send_request(
            method="GET",
            endpoint_url=self.__API_ENDPOINT_STREAM_URL,
            headers=request_headers
        )
        if response.status_code == 200:
            try:
                response_data = response.json()

                return [StreamResponse(data=stream, links=stream.get('links', None)) for stream in response_data['data']]
            except Exception:
                raise AttributeError('Unexpected Get Streams API Response.')
        elif response.status_code == 403:
            raise ValueError('Factiva API-Key does not exist or inactive.')
        else:
            raise RuntimeError('Unexpected Get Streams API Error')

    def fetch_credentials(self) -> dict:
        """Fetch the current headers and uri (v1 or v2).

        Then it is executed a call to an api with the given header and uri
        The result is a json with the streaming credentials for Pubsub

        Returns
        -------
        Json object -> credentials in json format obtained
        from Account Streams Credentials endpoint

        Raises
        ------
        ValueError: When the credentials are not valid
        RuntimeError: When API request returns unexpected error

        """
        headers = self.get_authentication_headers()
        uri = self.get_uri_context()
        response = api_send_request(
            method="GET",
            endpoint_url='{}{}'.format(uri, self.__API_ACCOUNT_STREAM_CREDENTIALS_BASEPATH),
            headers=headers
        )

        if response.status_code == 401:
            message = '''
                Extraction API authentication failed for given
                credentials header:{}
                '''.format(headers)
            raise RuntimeError(message)
        try:
            streaming_credentials_string = response.json()['data']['attributes']['streaming_credentials']
        except KeyError:
            raise ValueError(
                '''
                Unable to find streaming credentials for given account
                '''
            )

        return json.loads(streaming_credentials_string)

    def get_uri_context(self) -> str:
        """Obtain the current headers, return the uri based on the headers used.

        Returns
        -------
        uri in string format which can be used for
        OAuth or User-Key Authentication

        Raises
        ------
        ValueError: When the headers doesn't have
            OAuth credentials or user key

        """
        headers = self.get_authentication_headers()
        if "Authorization" in headers:
            return self.__DEFAULT_HOST_DNA

        if 'user-key' in headers:
            return self.__DEFAULT_HOST_ALPHA

        msg = '''
        Could not determine user credentials:
        Must specify account credentials as user_key
        through env vars
        (see README.rst)
        '''
        raise ValueError(msg)

    def get_client_subscription(self) -> SubscriberClient:
        """Obtain the subscriber client for pubsub.

        The credentials are obtained from fetch_credentials() function
        These credentials are used to authenticate with Google services
        If all is correct, it will be created a SubscriberClient for Pubsub

        Returns
        -------
        SubscriberClient object from google cloud library used for Pubsub

        Raises
        ------
        RuntimeError: When the Pubsub client cannot be created

        """
        streaming_credentials = self.fetch_credentials()
        try:
            credentials = service_account.Credentials.from_service_account_info(
                streaming_credentials
            )

            return SubscriberClient(credentials=credentials)
        except Exception:
            raise RuntimeError(
                '''
                Something unexpected happened while creating Pubsub client
                '''
            )

    def get_authentication_headers(self) -> dict:
        """Obtain the current auhtentication headers.

        The headers are the user-key

        Returns
        -------
        Object with User Key

        Raises
        ------
        ValueError: When the Pubsub client cannot be created

        """
        if self.api_key:
            return {'user-key': self.api_key}

        msg = '''
            Could not find credentials:
            Must specify account credenstials as user_key
            (see README.rst)
        '''
        raise ValueError(msg)



"""Implement Stream Response Class."""


class StreamResponse:
    """Represent a Stream Response Class in string format.

    Parameters
    ----------
    response: dict
        represents a given stream response by its json representation

    Raises
    ------
    ValueError: When there is an empty response

    Examples
    --------
    Create a StreamResponse
        >>> result = StreamResponse(response)
        >>> print(result)

    type: stream
    id:
        dj-synhub-extraction-*********************************
    attributes:
        job_status: JOB_STATE_RUNNING

    relationships:
        subscriptions:
                data:
                        id: dj-synhub-extraction-*********************************-filtered-*****
                        type: Subscription

                data:
                        id: dj-synhub-extraction-*********************************-filtered-*****
                        type: Subscription

                data:
                        id: dj-synhub-extraction-*********************************-filtered-*****
                        type: Subscription

    links:
        self: https://api.dowjones.com/alpha/streams/dj-synhub-extraction-*********************************

    """
    stype = ''
    id = ''
    attributes = ''
    relationships = ''
    links = ''

    def __init__(self, data=None, links=None):
        """Construct class instance."""
        if not data:
            raise ValueError('Empty value for reponse')
        self.parse_data(data)
        self.parse_links(links)

    def parse_data(self, data):
        """Parse data if it exists.

        Parameters
        ----------
        data: dict
            dict which must contain data

        Raises
        ------
        ValueError: When there is empty data

        """
        self.stype = data['type']
        # pylint: disable=invalid-name
        self.id = data['id']
        self.attributes = self.parse_object(data['attributes'])
        self.relationships = self.parse_object(data['relationships'])

    def parse_links(self, data):
        """Parse of existing links.

        Parameters
        ----------
        data: dict
            dict which may contains links

        """
        if data and 'links' in data:
            self.links = self.parse_object(data['links'])

    def parse_object(self, data, level=2):
        """Parse object representation per level of identation.

        Parameters
        ----------
        data: dict
            representation of every key value pair
        level: int
            level of identation needed

        Returns
        -------
        String which contains all attributes and values
        with identation and spaces included

        """
        object_repr = ''
        idents = "\t" * level
        for index_k, index_v in data.items():
            if isinstance(index_v, dict):
                object_repr += f'{idents}{index_k}: \n{self.parse_object(index_v, level + 1)}\n'
            elif isinstance(index_v, list):
                for att in index_v:
                    object_repr += f'{idents}{index_k}: \n{self.parse_object(att, level + 1)}\n'
            else:
                object_repr += f'{idents}{index_k}: {index_v}\n'

        return object_repr

    def __repr__(self):
        """Return the repr in the class intance."""
        return '''
            type: {}
            id: {}
            attributes: \n{}
            relationships: \n{}
            links: \n{}
            '''.format(
            self.stype,
            self.id,
            self.attributes,
            self.relationships,
            self.links,
        )



"""Implement core capabilities for classes."""
from .__version__ import __version__

from .apikeyuser import (APIKeyUser)

from .streamuser import (StreamUser)

from .stream_response import (StreamResponse)

__all__ = [APIKeyUser, StreamUser, StreamResponse, 'const', 'dicts', ]
version = __version__



"""Define library's constant literals."""
from .errors import *

API_HOST = 'https://api.dowjones.com'
API_ACCOUNT_OAUTH2_HOST = 'https://accounts.dowjones.com/oauth2/v1/token'

API_ACCOUNT_BASEPATH = '/alpha/accounts'
ALPHA_BASEPATH = '/alpha'
DNA_BASEPATH = '/dna'

# Snapshots
API_SNAPSHOTS_BASEPATH = '/alpha/extractions/documents'
API_EXPLAIN_SUFFIX = '/_explain'
API_ANALYTICS_BASEPATH = '/alpha/analytics'
API_EXTRACTIONS_BASEPATH = '/alpha/extractions'

API_SNAPSHOTS_TAXONOMY_BASEPATH = '/alpha/taxonomies'
API_SNAPSHOTS_COMPANIES_BASEPATH = '/alpha/companies'

# ANALYTICS
API_AVRO_FORMAT = 'avro'
API_CSV_FORMAT = 'csv'
API_JSON_FORMAT = 'json'
API_EXTRACTION_FILE_FORMATS = [API_AVRO_FORMAT, API_JSON_FORMAT, API_CSV_FORMAT]

API_DAY_PERIOD = 'DAY'
API_MONTH_PERIOD = 'MONTH'
API_YEAR_PERIOD = 'YEAR'
API_DATETIME_PERIODS = [API_DAY_PERIOD, API_MONTH_PERIOD, API_YEAR_PERIOD]

API_PUBLICATION_DATETIME_FIELD = 'publication_datetime'
API_MODIFICATION_DATETIME_FIELD = 'modification_datetime'
API_INGESTION_DATETIME_FIELD = 'ingestion_datetime'
API_DATETIME_FIELDS = [API_PUBLICATION_DATETIME_FIELD, API_MODIFICATION_DATETIME_FIELD, API_INGESTION_DATETIME_FIELD]

# Streams
API_STREAMS_BASEPATH = '/alpha/streams'
API_ACCOUNT_STREAM_CREDENTIALS_BASEPATH = '/accounts/streaming-credentials'
DOC_COUNT_EXCEEDED = "DOC_COUNT_EXCEEDED"
CHECK_EXCEEDED_WAIT_SPACING = 300
PUBSUB_MESSAGES_WAIT_SPACING = 10

# API STATES
API_JOB_CREATED_STATE = 'JOB_CREATED'
API_JOB_QUEUED_STATE = 'JOB_QUEUED'
API_JOB_PENDING_STATE = 'JOB_STATE_PENDING'
API_JOB_VALIDATING_STATE = 'JOB_VALIDATING'
API_JOB_STATE_VALIDATING = 'JOB_STATE_VALIDATING'
API_JOB_RUNNING_STATE = 'JOB_STATE_RUNNING'
API_JOB_DONE_STATE = 'JOB_STATE_DONE'
API_JOB_FAILED_STATE = 'JOB_STATE_FAILED'
API_JOB_CANCELLED_STATE = 'JOB_STATE_CANCELLED'

API_JOB_EXPECTED_STATES = [API_JOB_CREATED_STATE, API_JOB_QUEUED_STATE, API_JOB_PENDING_STATE,
                           API_JOB_VALIDATING_STATE,
                           API_JOB_STATE_VALIDATING, API_JOB_RUNNING_STATE, API_JOB_DONE_STATE, API_JOB_FAILED_STATE,
                           API_JOB_CANCELLED_STATE]

API_JOB_ACTIVE_WAIT_SPACING = 10



"""Implement different helper functions."""
import os
import json
import requests
from factiva.core import const


def load_environment_value(config_key):
    """Obtain a environmental variable."""
    tmp_val = os.getenv(config_key, None)
    if tmp_val is None:
        raise Exception("Environment Variable {} not found!".format(config_key))
    return tmp_val


def mask_string(raw_str, right_padding=4):
    """Mask a string with a default value of 4."""
    return raw_str[-right_padding:].rjust(len(raw_str), '*')


def send_get_request(endpoint_url=const.API_HOST, headers=None, qs_params=None):
    """Send get request."""
    if (qs_params is not None) and (not isinstance(qs_params, dict)):
        raise ValueError('Unexpected qs_params value')
    return requests.get(endpoint_url, headers=headers, params=qs_params)


def send_post_request(endpoint_url=const.API_HOST, headers=None, payload=None):
    """Send post request."""
    if payload is not None:
        if isinstance(payload, dict):
            payload_str = json.dumps(payload)
        elif isinstance(payload, str):
            payload_str = payload
        else:
            raise ValueError('Unexpected payload value')
        return requests.post(endpoint_url, headers=headers, data=payload_str)

    return requests.post(endpoint_url, headers=headers)


def api_send_request(method='GET', endpoint_url=const.API_HOST, headers=None, payload=None, qs_params=None):
    """Send a generic request to a certain API end point."""
    if headers is None:
        raise ValueError('Heders for Factiva requests cannot be empty')

    if not isinstance(headers, dict):
        raise ValueError('Unexpected headers value')

    try:
        if method == 'GET':
            response = send_get_request(endpoint_url=endpoint_url, headers=headers, qs_params=qs_params)

        elif method == 'POST':
            response = send_post_request(endpoint_url=endpoint_url, headers=headers, payload=payload)

        elif method == 'DELETE':
            response = requests.delete(endpoint_url, headers=headers)

        else:
            raise ValueError('Unexpected method value')

    except Exception:
        raise RuntimeError('API Request failed. Unspecified Error.')

    return response


def validate_type(var_to_validate, expected_type, error_message):
    """Validate a given type."""
    if not isinstance(var_to_validate, expected_type):
        raise ValueError(error_message)


def flatten_dict(multi_level_dict):
    """Flatten a dictionary."""
    flattened_dict = {}
    for entry in multi_level_dict:
        if isinstance(multi_level_dict[entry], dict):
            new_elements = flatten_dict(multi_level_dict[entry])
            flattened_dict.update(new_elements)
        else:
            flattened_dict[entry] = multi_level_dict[entry]

    return flattened_dict


def load_generic_env_variable(variable, env_var):
    """Load generic env variable."""
    if not variable:
        try:
            return load_environment_value(f'{env_var}')
        except Exception:
            return variable

    return variable



"""Factiva Core API Key User Class."""

import requests
import pandas as pd

from factiva.helper import load_environment_value, mask_string, api_send_request, flatten_dict
from factiva.core import const


class APIKeyUser:  # TODO: Create a DJUserBase class that defines root properties for all user types, and inherit here.
    """Class that represents an API user. This entity is identifiable by an API-Key.

    Parameters
    ----------
    api_key : str
        String containing the 32-character long APi Key. If not provided, the
        constructor will try to obtain its value from the FACTIVA_APIKEY
        environment variable.
    request_info : boolean, optional (Default: False)
        Indicates if user data has to be pulled from the server. This operation
        fills account detail properties along with maximum, used and remaining
        values. It may take several seconds to complete.

    See Also
    --------
    UserOAuth: API user that follows the OAuth guidelines.

    Examples
    --------
    Creating a new API user providing the API-Key explicitly and requesting to retrieve the latest account details:
        >>> aku = APIKeyUser('abcd1234abcd1234abcd1234abcd1234', request_info=True)
        >>> print(aku)
        (class 'factiva.core.apikeyuser.APIKeyUser')
            api_key = ****************************1234
            account_name = Demo Account
            account_type = account_with_limits
            active_products = Snapshots
            max_allowed_concurrent_extractions = 2
            max_allowed_extracted_documents = 100000
            max_allowed_extractions = 10
            total_downloaded_bytes = 12345678
            total_extracted_documents = 5500
            total_extractions = 2
            total_stream_subscriptions = 2
            total_stream_topics = 1
            remaining_documents = 94500
            remaining_extractions = 8
    Creating a new instance taking the api-key value from the environment varaible FACTIVA_APIKEY, and not requesting any account details.
        >>> aku = APIKeyUser()
        >>> print(aku)
        (class 'factiva.core.apikeyuser.APIKeyUser')
            api_key = ****************************1234
            account_name =
            account_type =
            active_products =
            max_allowed_concurrent_extractions = 0
            max_allowed_extracted_documents = 0
            max_allowed_extractions = 0
            total_downloaded_bytes = 0
            total_extracted_documents = 0
            total_extractions = 0
            total_stream_subscriptions = 0
            total_stream_topics = 0
            remaining_documents = 0
            remaining_extractions = 0

    """
    # pylint: disable=too-many-instance-attributes
    # Twelve is reasonable in this case.

    __API_ENDPOINT_BASEURL = f'{const.API_HOST}{const.API_ACCOUNT_BASEPATH}/'

    api_key = ''
    account_name = ''
    account_type = ''
    active_products = ''
    max_allowed_concurrent_extractions = 0
    max_allowed_extracted_documents = 0
    max_allowed_extractions = 0
    currently_running_extractions = 0
    total_downloaded_bytes = 0
    total_extracted_documents = 0
    total_extractions = 0
    total_stream_subscriptions = 0
    total_stream_topics = 0
    enabled_company_identifiers = []

    def __init__(
        self,
        api_key=None,
        request_info=False
    ):
        """Construct the instance of the class."""
        if api_key is None:
            try:
                api_key = load_environment_value('FACTIVA_APIKEY')
            except Exception:
                raise ValueError('Factiva API-Key not provided and environment variable FACTIVA_APIKEY not set.')

        if len(api_key) != 32:
            raise ValueError('Factiva API-Key has the wrong length')

        self.api_key = api_key

        if request_info is True:
            self.get_info()
        else:
            self.account_name = ''
            self.account_type = ''
            self.active_products = ''
            self.max_allowed_concurrent_extractions = 0
            self.max_allowed_extracted_documents = 0
            self.max_allowed_extractions = 0
            self.currently_running_extractions = 0
            self.total_downloaded_bytes = 0
            self.total_extracted_documents = 0
            self.total_extractions = 0
            self.total_stream_subscriptions = 0
            self.total_stream_topics = 0
            self.enabled_company_identifiers = []

    @property
    def remaining_extractions(self):
        """Account remaining extractions."""
        return self.max_allowed_extractions - self.total_extractions

    @property
    def remaining_documents(self):
        """Account remaining documents."""
        return self.max_allowed_extracted_documents - self.total_extracted_documents

    # [MB] TODO: Please move this method to factiva-news/snapshot/snapshot.py
    # The number of extractions is OK under this class for account management
    # information (get_info), but the extraction list belongs to the Snapshot class.
    def get_extractions(self) -> pd.DataFrame:
        """Request a list of the extractions of the account.

        Returns
        -------
        Dataframe containing the information about the account extractions

        Raises
        ------
        - ValueError when the API Key provided is not valid
        - RuntimeError when the API returns an unexpected error

        """
        endpoint = f'{const.API_HOST}{const.API_EXTRACTIONS_BASEPATH}'

        headers_dict = {'user-key': self.api_key}

        response = api_send_request(method='GET', endpoint_url=endpoint, headers=headers_dict)

        if response.status_code != 200:
            if response.status_code == 403:
                raise ValueError('Factiva API-Key does not exist or inactive.')

            raise RuntimeError(f'Unexpected API Error with message: {response.text}')

        response_data = response.json()

        extraction_list = [flatten_dict(extraction) for extraction in response_data['data']]

        return pd.DataFrame(extraction_list)

    def get_info(self):
        """Request the account details to the Factiva Account API Endpoint.

        This operation can take several seconds to complete.

        Returns
        -------
        True if the operation was completed successfully. All returned values
        are set to the object's properties directly.

        Examples
        --------
        >>> aku = APIKeyUser('abcd1234abcd1234abcd1234abcd1234')
        >>> aku
        (class 'factiva.core.apikeyuser.APIKeyUser')
            api_key = ****************************1234
            account_name =
            account_type =
            active_products =
            max_allowed_concurrent_extractions = 0
            max_allowed_extracted_documents = 0
            max_allowed_extractions = 0
            total_downloaded_bytes = 0
            total_extracted_documents = 0
            total_extractions = 0
            total_stream_subscriptions = 0
            total_stream_topics = 0
            remaining_documents = 0
            remaining_extractions = 0
        >>> aku.get_info()
        >>> aku
        (class 'factiva.core.apikeyuser.APIKeyUser')
            api_key = ****************************1234
            account_name = Demo Account
            account_type = account_with_limits
            active_products = Snapshots
            max_allowed_concurrent_extractions = 2
            max_allowed_extracted_documents = 100000
            max_allowed_extractions = 10
            total_downloaded_bytes = 12345678
            total_extracted_documents = 5500
            total_extractions = 2
            total_stream_subscriptions = 2
            total_stream_topics = 1
            remaining_documents = 94500
            remaining_extractions = 8

        """
        account_endpoint = f'{self.__API_ENDPOINT_BASEURL}{self.api_key}'
        req_head = {'user-key': self.api_key}
        resp = requests.get(account_endpoint, headers=req_head)  # TODO: Consider processing all GET/POST requests in a separate class/module
        if resp.status_code == 200:
            try:
                resp_obj = eval(resp.text)
                self.account_name = resp_obj['data']['attributes']['name']
                self.account_type = resp_obj['data']['type']
                self.active_products = resp_obj['data']['attributes']['products']
                self.max_allowed_concurrent_extractions = resp_obj['data']['attributes']['max_allowed_concurrent_extracts']
                self.max_allowed_extracted_documents = resp_obj['data']['attributes']['max_allowed_document_extracts']
                self.max_allowed_extractions = resp_obj['data']['attributes']['max_allowed_extracts']
                self.currently_running_extractions = resp_obj['data']['attributes']['cnt_curr_ext']
                self.total_downloaded_bytes = resp_obj['data']['attributes']['current_downloaded_amount']
                self.total_extracted_documents = resp_obj['data']['attributes']['tot_document_extracts']
                self.total_extractions = resp_obj['data']['attributes']['tot_extracts']
                self.total_stream_subscriptions = resp_obj['data']['attributes']['tot_subscriptions']
                self.total_stream_topics = resp_obj['data']['attributes']['tot_topics']
                self.enabled_company_identifiers = resp_obj['data']['attributes']['enabled_company_identifiers']
            except Exception:
                raise AttributeError('Unexpected Account Information API Response.')
        elif resp.status_code == 403:
            raise ValueError('Factiva API-Key does not exist or inactive.')
        else:
            raise RuntimeError('Unexpected Account Information API Error')
        return True

    def __print_property__(self, property_value) -> str:
        if type(property_value) == int:
            pval = f'{property_value:,d}'
        else:
            pval = property_value
        return pval

    def __repr__(self):
        """Return a string representation of the object."""
        return self.__str__()

    def __str__(self, detailed=True, prefix='  |-', root_prefix=''):
        pprop = self.__dict__.copy()
        del pprop['api_key']
        masked_key = mask_string(self.__dict__['api_key'])

        ret_val = f'{root_prefix}{str(self.__class__)}\n'
        ret_val += f'{prefix}api_key = {masked_key}\n'
        if detailed:
            ret_val += '\n'.join((f'{prefix}{item} = {self.__print_property__(pprop[item])}' for item in pprop))
            ret_val += f'\n{prefix}remaining_documents = {self.__print_property__(self.remaining_documents)}\n'
            ret_val += f'{prefix}remaining_extractions = {self.__print_property__(self.remaining_extractions)}\n'
        else:
            ret_val += f'{prefix}...'
        return ret_val

    @staticmethod
    def create_api_user(api_user, request_userinfo):
        """Determine the way to initialize an api key user according to the type of parameter provided.

        Parameters
        ----------
        api_user: None, str or APIKeyUser.
                Source to create an instance of APIKeyUser
        request_userinfo
                Indicates if user data has to be pulled from the server

        Returns
        -------
        APIKeyUser instance accordingly:
            - When None is passed, APIKeyUser instance using credentials from ENV variables
            - When str is passed, APIKeyUser instance using the provided parameter as credentials
            - When ApiKeyUser is passed, this same instance is returned

        Raises
        ------
            RuntimeError: When an APIKeyUser instance cannot be created using the provided argument

        """
        if isinstance(api_user, APIKeyUser):
            return api_user

        if isinstance(api_user, str):
            try:
                return APIKeyUser(api_user, request_info=request_userinfo)
            except Exception:
                raise RuntimeError("User cannot be obtained from the provided key.")

        if api_user is None:
            try:
                return APIKeyUser(request_info=request_userinfo)
            except Exception:
                raise RuntimeError("User cannot be obtained from ENV variables")

        raise RuntimeError("Unexpected api_user value")
